/*************************************************************************************      
 *                                                               Micha Chodkiewicz  *
 * Wzorzec:         fabryka abstrakcyjna (abstract factory)                          *
 *                                                                                   *
 * Zastosowanie:    Wzorzec ten umoliwia atwe wprowadzanie zmian w zakresie klas   *
 *                  uywanych w kodzie, co umoliwia pisanie kodu, np: na kilka      * 
 *                  platform jednoczenie, czy atwiejsze zmienianie kodu wewntrz   *
 *                  klasy bibliotecznej.                                             *
 *                                                                                   *   
 *   Opis:          Fabryka abstrakcyjna, zwykle tworzona jako interfejs, w jzyku   *
 *                  c++ jest implementowana za pomoc klasy abstrakcyjnej. Klasa ta  *
 *                  ma wygld typowej fabryki, jednak jako typy zwracane przez       *
 *                  wirtualne metody fabrykujce ma zadeklarowane klasy abstrakcyjne *
 *                  reprezentujce interfejsy obiektw zwracanych przez klasy        *
 *                  pochodne. W kodzie programu implementuje si dla konkretnych     *
 *                  zastosowa klasy pochodne od fabryki abstrakcyjnej wytwarzajce  *
 *                  obiekty odpowiednie dla danego systemu czy pliku                 *
 *                  konfiguracyjnego. Dziki temu, e najczciej fabryka jest       *
 *                  tworzona w programie tylko raz, przy zmianie zachowania obiektw *
 *                  wystarczy zmodyfikowa t linijk w kodzie, w ktrej nastpuje   *
 *                  przypisanie konkretnej fabryki.                                  *
 *                                                                                   *
 *   Przykadowy schemat (schemat opisuje inny przypadek ni kod, dla urozmaicenia): *
 *   Prezentowany jest przypadek programowania dla kilku systemw operacyjnych       *
 *                                                                                   *
 *                  Odczyt schematu:                                                 *
 *                  [Element na pocztku strzaki]                                   *
 *                  (d) dziedziczy po                                                *
 *                  (w) zawiera wskanik/wskaniki na                                *
 *                  (t) tworzy obiekty klasy                                         *
 *                  [Element na ktry wskazuje strzaka]                             *
 *               _______________               ______________________                *
 *               | KlasaUsera  |-------------->|FabrykaAbstrakcyjna |                *
 *               |_____________|            (w)|____________________|                *
 *               |_____________|               |____________________|                *
 *               |_____________|               | A stworzObiekt()=0 |                *
 *                     |                       |____________________|                *
 *                     |                          /|\(d) /|\(d)  /|\(d)              *
 *                     |             ______________|      |       |________          *
 *                     |     _______|__________ __________|_______ ________|_________*
 *                     |     |FabrykaUnixowa  | |FabrykaDosowa   | |FabrykaWindows  |*
 *    -----------------/\----|________________| |________________| |________________|*
 *    |                |     |________________| |________________| |________________|*
 *    |                |     |A stworzObiekt()| |A stworzObiekt()| |A stworzObiekt()|*
 *    |                |(w)  |________________| |________________| |________________|*
 *    |    ___________\|/_____                           |                  |        *
 *    |    | A (klasa abstr) |                           |                  |        *
 *    |    |_________________|                           |                  |        *
 *    |    |_________________|                           |                  |        *
 *    |    | zareaguj()      |                           |                  |        *
 *    |    |_________________|                           |                  |        *
 *    |     /|\(d)  /|\(d)  /|\(d)                       |                  |        *
 *    |(t)___|       |       |_______                    |                  |        *
 *   \|/__|______ ___|________ _____|______              |                  |        *
 *   | AUnix    | | ADos     | | AWindows |              |                  |        *
 *   |__________| |__________| |__________|   (t)        |                  |        *
 *   |__________| |__________| |__________|<-------------/\-----------------|        *
 *   |zareaguj()| |zareaguj()| |zareaguj()|              |                           *
 *   |__________| |__________| |__________|              |                           *
 *                     /|\(t)                            |                           *
 *                      |________________________________|                           *
 *                                                                                   *
 *************************************************************************************/

/* Przykadowy kod. Przedstawia on rysowanie obiektw w zalenoci od
   szybkoci procesora */
#include <iostream>//Biblioteka potrzebna w tym przykadzie
using namespace std;// dla wygody, eby nie pisa wszdzie std

// Abstrakcyjny model obiektu graficznego
class ObiektGraficznyAbstrakcyjny {
   protected:   
      float x_;// pooenie na ekranie w osi x
      float y_;// pooenie na ekranie w osi y
   public:
         virtual void rysujSie()=0;// metoda do rysowania obiekt
         // ponisze metody s wirtualne, gdy klasy pochodne nie musz koniecznie w ten sposb przechowywa
         // pooenia obiektu, jednak maj ju ciaa, gdy to bdzie najczstszy przypadek

         // metoda do ustalania pooenia obiektu
         virtual void ustawPozycje(float x, float y){
            x_=x;
            y_=y;         
         }
         // metody do pobierania pooenia obiektu
         virtual float podajPozX() {return x_;}
         virtual float podajPozY() {return y_;}
         ObiektGraficznyAbstrakcyjny (float x, float y) : x_(x), y_(y){}
         virtual ~ObiektGraficznyAbstrakcyjny() {}//wirtualny destruktor, bo s wirtualne funkcje
};

// Abstrakcyjna fabryka obiektw, suca za interfejs
class FabrykaObiektowGraficznychAbstrakcyjna {
   public:
         // metoda do tworzenia instancji obiektw
         virtual ObiektGraficznyAbstrakcyjny * utworzObiekt(float x, float y)=0;
         virtual ~FabrykaObiektowGraficznychAbstrakcyjna(){}//wirtualny destruktor, bo s wirtualne funkcje
}; 

// Konkretny model obiektu graficznego, dla wolniejszych komputerw
class ObiektGraficznyProsty : public ObiektGraficznyAbstrakcyjny {
   public:
         // funkcja rysujSie jest mocno okrojona, gdy to tylko przykad zastosowania
         void rysujSie(){
            cout<<"malo wymagajacy model jest rysowany na pozycji: "<<x_<<", "<<y_<<endl;
         }
         ObiektGraficznyProsty (float x, float y):ObiektGraficznyAbstrakcyjny(x,y){}
};

// Konkretny model obiektu graficznego, dla szybszych komputerw
class ObiektGraficznyTrudny : public ObiektGraficznyAbstrakcyjny {
   public:
         // funkcja rysujSie jest mocno okrojona, gdy to tylko przykad zastosowania
         void rysujSie(){
            cout<<"BARDZO WYMAGAJACY MODEL JEST RYSOWANY NA POZYCJI: "<<x_<<", "<<y_<<endl;
         }
         ObiektGraficznyTrudny (float x, float y):ObiektGraficznyAbstrakcyjny(x,y){}
};

//Konkretna fabryka obiektw graficznych, dla wolnych komputerw
class FabrykaObiektowGraficznychProstych : public FabrykaObiektowGraficznychAbstrakcyjna {
   public:
      // skonkretyzowanie metody
      ObiektGraficznyAbstrakcyjny * utworzObiekt(float x, float y) {
            return new ObiektGraficznyProsty(x,y);      
      }
};

//Konkretna fabryka obiektw graficznych, dla szybkich komputerw
class FabrykaObiektowGraficznychTrudnych : public FabrykaObiektowGraficznychAbstrakcyjna {
   public:
      // skonkretyzowanie metody
      ObiektGraficznyAbstrakcyjny * utworzObiekt(float x, float y) {
            return new ObiektGraficznyTrudny(x,y);      
      }
};

// Klasa korzystajca z fabryk
class KlasaUzytkowa {
   private:
      static const int ILE_OBIEKTOW=5;//ile obiektow graficznych naraz na ekranie maksymalnie
      // Wskaznik na fabryke obiektow, tu bedzie podany wskaznik do konkretnej fabryki 
      FabrykaObiektowGraficznychAbstrakcyjna * fabryka_;
      ObiektGraficznyAbstrakcyjny * pulaObiektow_[ILE_OBIEKTOW]; // Tablica obiektow do rysowania
   public:
      // dla zapeniania tablicy obiektw, zmienna nowa partia mwi czy zmienia obiekty,
      // czy te tworzy nowe od poczatku
      void pobierzNowe(bool nowaPartia){
         for (int licznik=0; licznik<5; ++licznik){
            ObiektGraficznyAbstrakcyjny * temp=pulaObiektow_[licznik];         
            if (temp!=NULL){// odtworzenie obiektow w nowych warunkach
                  pulaObiektow_[licznik]=fabryka_->utworzObiekt(temp->podajPozX(), temp->podajPozY());   
                  delete temp;          
            }
            else if (nowaPartia){
               pulaObiektow_[licznik]=fabryka_->utworzObiekt(licznik, ILE_OBIEKTOW - licznik);   
            }            
         }      
      }
      // dla ustawienia rda obiektw
      void ustawFabryke(FabrykaObiektowGraficznychAbstrakcyjna * nowa_fabryka) {
            delete fabryka_;// usunicie starej fabryki            
            fabryka_=nowa_fabryka;   // ustawienie nowej fabryki
            pobierzNowe(false);//eby od razu zmieni typ przechowywanych obiektw   
      } 
      // dla rysowania obiektw
      void rysujObiekty(){
         for (int licznik=0; licznik<ILE_OBIEKTOW; ++licznik){
            if (pulaObiektow_[licznik]!=NULL){
                  pulaObiektow_[licznik]->rysujSie();      
            }         
         }      
      }
      // konstruktor
      KlasaUzytkowa (FabrykaObiektowGraficznychAbstrakcyjna *nowa_fabryka){
            fabryka_=nowa_fabryka;      
      }
      // destruktor
      ~KlasaUzytkowa(){
         for (int licznik=0; licznik<5; ++licznik){
            delete pulaObiektow_[licznik];// zwolnienie puli obiektw      
         }         
         delete fabryka_;// zwolnienie fabryki
      }      
};

//Program testowy
int main(){
      cout<<"..:: Uruchamianie programu w trybie wysokiej jakosci grafiki ::.."<<endl;
      // Na pocztek ustawiany bardziej wymagajcy typ obiektu graficznego
      KlasaUzytkowa * testowa=new KlasaUzytkowa(new FabrykaObiektowGraficznychTrudnych);
      testowa->pobierzNowe(true);// Ustawianie przykadowych obiektw
      cout<<"..:: Rysowanie przykladowych obiektow ::.."<<endl;
      testowa->rysujObiekty();
      //{...} Tutaj okazao si, e niestety maszyna jest za wolna
      cout<<"..:: Niestety maszyna jest za wolna, przelaczanie w tryb niskiej jakosci::.."<<endl;
      testowa->ustawFabryke(new FabrykaObiektowGraficznychProstych());
      cout<<"..:: Rysowanie przykladowych obiektow ::.."<<endl;
      testowa->rysujObiekty();
      cout<<"..:: Koniec programu ::.."<<endl;
}
